# -*- coding: utf-8 -*-

from copy import copy, deepcopy
from openpyxl.worksheet.cell_range import CellRange, MultiCellRange
from .merger import MergeMixin


class MergerMixin():

    @property
    def to_merge(self):
        return bool(self._merge_list)

class CellMerge(MergeMixin):

    def __init__(self, cell_range, merger):
        self.merger = merger
        self.set_range()
        self._first_row = cell_range.min_row
        self._last_row = cell_range.max_row
        self._first_col = cell_range.min_col
        self._last_col = cell_range.max_col

    def new_range(self):
        if self.start_wtrowx==self.end_wtrowx and self.start_wtcolx==self.end_wtcolx:
            return
        range = CellRange(None, self.start_wtcolx, self.start_wtrowx, self.end_wtcolx, self.end_wtrowx)
        self.merger.add_new_range(range)

class CellMerger(MergerMixin):

    def __init__(self, sheet):
        self.range_list = []
        self._merge_list = []
        self.get_merge_list(sheet)

    def get_merge_list(self, sheet):
        for range in sheet.merged_cells:
            _merge = CellMerge(range, self)
            self._merge_list.append(_merge)

    def add_new_range(self, range):
        self.range_list.append(range)

    def merge_cell(self, rdrowx, rdcolx, wtrowx, wtcolx):
        for _merge in self._merge_list:
            is_in_range = _merge.merge_cell(rdrowx, rdcolx, wtrowx, wtcolx)
            if is_in_range:
                break

    def collect_range(self, wtsheet):
        for _merge in self._merge_list:
            _merge.collect_range()
        for range in self.range_list:
            wtsheet.merged_cells.add(range)
        self.range_list.clear()

class DataValidation(MergeMixin):

    def __init__(self, cell_range, merger, dv_key):
        self.dv_key = dv_key
        self.merger = merger
        self.set_range()
        self._first_row = cell_range.min_row
        self._last_row = cell_range.max_row
        self._first_col = cell_range.min_col
        self._last_col = cell_range.max_col

    def new_range(self):
        if self.start_wtrowx==-1:
            return
        range = CellRange(None, self.start_wtcolx, self.start_wtrowx, self.end_wtcolx, self.end_wtrowx)
        self.merger.add_new_range(self.dv_key, range)

class DvMerger(MergerMixin):

    def __init__(self, sheet):
        self.dv_map = {}
        self.dv_copy_map = {}
        self._merge_list = []
        self.get_merge_list(sheet)

    def get_merge_list(self, rdsheet):
        for index,dv in enumerate(rdsheet.data_validations.dataValidation):
            self.dv_map[index] = dv
            for crange in dv.ranges:
                _merge = DataValidation(crange, self, index)
                self._merge_list.append(_merge)

    def add_new_range(self, dv_key, range):
        dv_copy = self.dv_copy_map.get(dv_key)
        if not dv_copy:
            dv_copy = copy(self.dv_map[dv_key])
            dv_copy.ranges = MultiCellRange()
            self.dv_copy_map[dv_key] = dv_copy
        dv_copy.ranges.add(range)

    def merge_cell(self, rdrowx, rdcolx, wtrowx, wtcolx):
        for _merge in self._merge_list:
            is_in_range = _merge.merge_cell(rdrowx, rdcolx, wtrowx, wtcolx)
            if is_in_range:
                break

    def collect_range(self, wtsheet):
        for _merge in self._merge_list:
            _merge.collect_range()
        for key, dv in self.dv_copy_map.items():
            wtsheet.data_validations.append(dv)
        self.dv_copy_map.clear()

from collections import defaultdict
class ImageMerge(MergeMixin):

    def __init__(self, image, merger, image_count_dict):
        self.merger = merger
        self.image = image
        self.set_range()
        self.image_copy_map = {}
        self.image_ref_map = {}
        _from = image.anchor._from
        _to = image.anchor.to
        self._first_row = rlo = _from.row + 1
        self._first_col = clo = _from.col + 1
        self._last_row = rhi = _to.row + 1
        self._last_col = chi = _to.col + 1
        _top_left = (rlo, clo)
        count = image_count_dict[_top_left]
        image_count_dict[_top_left] += 1
        self.image_key = (rlo, clo, count)

    def new_range(self):
        if self.start_wtrowx==-1:
            return
        image = deepcopy(self.image)
        _from = image.anchor._from
        _to = image.anchor.to
        _from.row = self.start_wtrowx - 1
        _from.col = self.start_wtcolx - 1
        _to.row = self.end_wtrowx - 1
        _to.col = self.end_wtcolx - 1
        self.image_copy_map[(self.start_wtrowx, self.start_wtcolx)] = image

    def set_image_ref(self, image_ref):
        if image_ref.image:
            self.image_ref_map[image_ref.wt_top_left] = image_ref.image

    def collect_range(self):
        self.new_range()
        self.set_range()
        for key,image in self.image_copy_map.items():
            ref = self.image_ref_map.get(key)
            if ref:
                image.ref = ref
            self.merger.add_image(image)
        self.image_copy_map.clear()
        self.image_ref_map.clear()

class ImageMerger(MergerMixin):

    def __init__(self, sheet):
        self.images = []
        self._merge_map = {}
        self._merge_list = []
        self.max_row = 0
        self.max_col = 0
        self.get_merge_list(sheet)

    def get_merge_list(self, rdsheet):
        image_count_dict = defaultdict(int)
        for image in rdsheet._images:
            _merge = ImageMerge(image, self, image_count_dict)
            self._merge_map[_merge.image_key] = _merge
            self._merge_list.append(_merge)
            self.max_row = max(self.max_row, _merge._last_row)
            self.max_col = max(self.max_col, _merge._last_col)

    def add_image(self, image):
        self.images.append(image)

    def set_image_ref(self, image_ref):
        _merge = self._merge_map.get(image_ref.image_key)
        if not _merge:
            return
        _merge.set_image_ref(image_ref)

    def merge_cell(self, rdrowx, rdcolx, wtrowx, wtcolx):
        for _merge in self._merge_list:
            _merge.merge_cell(rdrowx, rdcolx, wtrowx, wtcolx)

    def collect_range(self, wtsheet):
        for _merge in self._merge_list:
            _merge.collect_range()
        wtsheet._images = self.images
        self.images = []

class AutoFilter(MergeMixin):

    def __init__(self, rdsheet):
        if not rdsheet.auto_filter.ref:
            self.to_merge = False
            return
        self.to_merge = True
        self.auto_filter = rdsheet.auto_filter
        self.set_range()
        cell_range = CellRange(rdsheet.auto_filter.ref)
        self._first_row = cell_range.min_row
        self._last_row = cell_range.max_row
        self._first_col = cell_range.min_col
        self._last_col = cell_range.max_col
        self.first_af = None

    def new_range(self):
        if self.start_wtrowx==-1:
            return
        if not self.first_af:
            self.first_af = CellRange(None, self.start_wtcolx, self.start_wtrowx,
                                      self.end_wtcolx, self.end_wtrowx)

    def collect_range(self, wtsheet):
        self.new_range()
        self.set_range()
        if wtsheet.auto_filter.ref:
            self.first_af = None
            return
        if self.first_af:
            wtsheet.auto_filter = copy(self.auto_filter)
            wtsheet.auto_filter.ref = self.first_af.coord
            self.first_af = None


class Merger:

    def __init__(self, rdsheet):
        cell_merger = CellMerger(rdsheet)
        dv_merger = DvMerger(rdsheet)
        self.image_merger = image_merger = ImageMerger(rdsheet)
        auto_filter = AutoFilter(rdsheet)
        _merger_list = [cell_merger, dv_merger, image_merger, auto_filter]
        self.merger_list = []
        for merger in _merger_list:
            if merger.to_merge:
                self.merger_list.append(merger)

    def merge_cell(self, rdrowx, rdcolx, wtrowx, wtcolx):
        for merger in self.merger_list:
            merger.merge_cell(rdrowx, rdcolx, wtrowx, wtcolx)

    def collect_range(self, wtsheet):
        for merger in self.merger_list:
            merger.collect_range(wtsheet)

    def set_image_ref(self, image_ref):
        self.image_merger.set_image_ref(image_ref)




# -*- coding: utf-8 -*-

import six
from copy import copy
from .utils import fix_test, tag_fix
from openpyxl.cell.text import RichText, Text

class RichTextHandler():

    def iter(self, rich_text, font):
        text_4_fix = self.text_4_fix(rich_text)
        if fix_test(text_4_fix):
            fixed = tag_fix(text_4_fix)
            #print(text_4_fix)
            #print(fixed)
            for i, segment in enumerate(rich_text):
                if i in fixed:
                    text = fixed[i]
                    if text == '':
                        continue
                    else:
                        yield text,segment[1],segment
        else:
            for segment in rich_text:
                yield segment[0],segment[1],segment

    @classmethod
    def rich_segment(self, text, font):
        return (text, font)

    @classmethod
    def text_content(cls, value):
        if isinstance(value, six.text_type):
            return value
        elif isinstance(value, list):
            x = []
            for text, font in value:
                x.append(text)
            return ''.join(x)

    @classmethod
    def text_4_fix(self, rich_text):
        text = []
        fmt = '___%d___'
        for i, segment in enumerate(rich_text):
            text.append(fmt % i)
            text.append(segment[0])
        return ''.join(text)

    @classmethod
    def rich_content(cls, value):
        segments = []
        for segment in value:
            if segment[0]:
                segments.append(segment)
        if segments:
            return segments
        else:
            return ''

    @classmethod
    def mid(cls, rich_text, head, tail):
        st = 0
        end = -1
        segments = []
        texts = []
        for index, segment in enumerate(rich_text):
            segment_text = segment[0]
            font = segment[1]
            l_text = len(segment_text)
            st = end + 1
            end += l_text
            if end < head:
                continue
            elif st <= head <= end:
                if end < tail:
                    text_st = head - st
                    text = segment_text[text_st:]
                    segments.append((text, font))
                    texts.append(text)
                else:
                    text_st = head - st
                    text_end = tail - st
                    text = segment_text[text_st:text_end + 1]
                    segments.append((text, font))
                    texts.append(text)
                    break
            elif end < tail:
                text = segment_text
                segments.append((text, font))
                texts.append(text)
            else:
                text_end = tail - st
                text = segment_text[:text_end + 1]
                segments.append((text, font))
                texts.append(text)
                break
        return segments, ''.join(texts)

class RichTextHandlerX():

    @classmethod
    def iter(self, rich_text, font):
        text_4_fix = self.text_4_fix(rich_text)
        if fix_test(text_4_fix):
            fixed = tag_fix(text_4_fix)
            #print(text_4_fix)
            #print(fixed)
            for i, segment in enumerate(rich_text.r):
                if i in fixed:
                    text = fixed[i]
                    if text == '':
                        continue
                    else:
                        segment_font = segment.font or font
                        yield text, segment_font, segment
        else:
            for segment in rich_text.r:
                segment_font = segment.font or font
                yield segment.text, segment_font, segment

    @classmethod
    def rich_segment(cls, text, font):
        return RichText(t=text, rPr=font)

    @classmethod
    def text_content(cls, value):
        return value

    @classmethod
    def text_4_fix(self, rich_text):
        text = []
        fmt = '___%d___'
        for i,segment in enumerate(rich_text.r):
            text.append(fmt % i)
            text.append(segment.text)
        return ''.join(text)

    @classmethod
    def rich_content(cls, value):
        if type(value) is Text:
            return value.content
        segments = []
        for segment in value:
            if segment.text:
                segments.append(segment)
        if segments:
            return Text(r=segments).content
        else:
            return ''

    @classmethod
    def mid(cls, rich_text, head, tail):
        st = 0
        end = -1
        segments = []
        texts = []
        for index, segment in enumerate(rich_text.r):
            l_text = len(segment.text)
            st = end + 1
            end += l_text
            if end < head:
                continue
            elif st <= head <= end:
                if end < tail:
                    text_st = head - st
                    segment_copy = copy(segment)
                    text = segment.text[text_st:]
                    segment_copy.text = text
                    segments.append(segment_copy)
                    texts.append(text)
                else:
                    text_st = head - st
                    text_end = tail - st
                    segment_copy = copy(segment)
                    text = segment.text[text_st:text_end+1]
                    segment_copy.text = text
                    segments.append(segment_copy)
                    texts.append(text)
                    break
            elif end < tail:
                segment_copy = copy(segment)
                text = segment.text
                #segment_copy.text = text
                segments.append(segment_copy)
                texts.append(text)
            else:
                text_end = tail - st
                segment_copy = copy(segment)
                text = segment.text[:text_end + 1]
                segment_copy.text = text
                segments.append(segment_copy)
                texts.append(text)
                break
        return Text(r=segments), ''.join(texts)


rich_handlerx = RichTextHandlerX()
rich_handler = RichTextHandler()



# -*- coding: utf-8 -*-

from .patchx import *
from openpyxl import load_workbook
from .basex import SheetBase, BookBase
from .writermixin import SheetMixin, BookMixin, Box
from .utils import tag_test, parse_cell_tag
from .xlnode import Tree, Row, Cell, EmptyCell, Node, create_cell
from .jinja import JinjaEnvx
from .nodemap import NodeMap
from .sheetresource import SheetResourceMap
from .richtexthandler import rich_handlerx
from .mergerx import Merger
from .config import config
from .celltag import CellTag

class SheetWriter(SheetBase, SheetMixin):

    def __init__(self, bookwriter, sheet_resource, sheet_name):
        self.workbook = bookwriter.workbook
        self.merger = sheet_resource.merger
        self.rdsheet = sheet_resource.rdsheet
        self.wtsheet = self.workbook.create_sheet(title=sheet_name)
        self.copy_sheet_settings()
        self.wtrows = set()
        self.wtcols = set()
        self.box = Box(0, 0)


class BookWriter(BookBase, BookMixin):
    sheet_writer_cls = SheetWriter

    def __init__(self, fname, debug=False):
        config.debug = debug
        self.load(fname)

    def load(self, fname):
        self.workbook = load_workbook(fname)
        self.font_map = {}
        self.node_map = NodeMap()
        self.jinja_env = JinjaEnvx(self.node_map)
        self.merger_cls = Merger
        self.sheet_writer_map = {}
        self.sheet_resource_map = SheetResourceMap(self, self.jinja_env)
        for index,rdsheet in enumerate(self.workbook.worksheets):
            self.sheet_resource_map.add(rdsheet, rdsheet.title, index)
            self.workbook.remove(rdsheet)

    def build(self, sheet, index, merger):
        tree = Tree(index, self.node_map)
        max_row = max(sheet.max_row, merger.image_merger.max_row)
        max_col = max(sheet.max_column, merger.image_merger.max_col)
        for rowx in range(1, max_row + 1):
            row_node = Row(rowx)
            tree.add_child(row_node)
            for colx in range(1, max_col + 1):
                sheet_cell = sheet._cells.get((rowx, colx))
                if not sheet_cell:
                    cell_node = EmptyCell(rowx, colx)
                    tree.add_child(cell_node)
                    continue
                cell_tag_map = None
                if sheet_cell.comment:
                    comment = sheet_cell.comment.text
                    if tag_test(comment):
                        _,cell_tag_map = parse_cell_tag(comment)
                value = sheet_cell._value
                data_type = sheet_cell.data_type
                rich_text = None
                if hasattr(value, 'rich') and value.rich:
                    rich_text = value.rich
                if data_type == 's':
                    if not tag_test(value):
                        cell_node = Cell(sheet_cell, rowx, colx, value, data_type)
                    else:
                        font = self.get_font(sheet_cell._style.fontId)
                        cell_node = create_cell(sheet_cell, rowx, colx, value, rich_text, data_type, font, rich_handlerx)
                else:
                    cell_node = Cell(sheet_cell, rowx, colx, value, data_type)
                if cell_tag_map:
                    cell_tag = CellTag(cell_tag_map)
                    cell_node.extend_cell_tag(cell_tag)
                    if colx==1:
                        row_node.cell_tag = cell_tag
                tree.add_child(cell_node)
        tree.add_child(Node())#
        return tree

    def cleanup_defined_names(self):
        sheet_cnt = len(self.workbook.worksheets)
        valid_names = []
        for d in self.workbook.defined_names.definedName:
            if d.localSheetId:
                if int(d.localSheetId) < sheet_cnt:
                    valid_names.append(d)
            else:
                valid_names.append(d)
        self.workbook.defined_names.definedName = valid_names

    def save(self, fname):
        if not self.workbook.active:
            self.workbook.active = 0
        self.cleanup_defined_names()
        self.workbook.save(fname)
        for sheet in self.workbook.worksheets:
            self.workbook.remove(sheet)
        self.sheet_writer_map.clear()



import os
from io import open
from setuptools import setup

CUR_DIR = os.path.abspath(os.path.dirname(__file__))
README = os.path.join(CUR_DIR, "README_EN.md")
with open(README, 'r', encoding='utf-8') as fd:
    long_description = fd.read()

setup(
    name = 'xltpl',
    version = "0.19",
    author = 'Zhang Yu',
    author_email = 'zhangyu836@gmail.com',
    url = 'https://github.com/zhangyu836/xltpl',
    packages = ['xltpl'],
	install_requires = ['xlrd >= 1.2.0', 'xlwt >= 1.3.0', 'openpyxl >= 2.6.0, <= 3.0.9', 'jinja2', 'six'],
    description = ( 'A python module to generate xls/x files from a xls/x template' ),
    long_description = long_description,
    long_description_content_type = "text/markdown",
    platforms = ["Any platform "],
    license = 'MIT',
    keywords = ['Excel', 'xls', 'xlsx', 'spreadsheet', 'workbook', 'template']
)



import copy
import xlrd
import xlwt
from openpyxl.utils.datetime import to_excel
from openpyxl.cell.cell import NUMERIC_TYPES, TIME_TYPES, STRING_TYPES
BOOL_TYPE = bool

def get_type(value):
    if isinstance(value, NUMERIC_TYPES):
        dt = xlrd.XL_CELL_NUMBER
    elif isinstance(value, STRING_TYPES):
        dt = xlrd.XL_CELL_TEXT
    elif isinstance(value, TIME_TYPES):
        dt = xlrd.XL_CELL_DATE
        return to_excel(value), dt
    elif isinstance(value, BOOL_TYPE):
        dt = xlrd.XL_CELL_BOOLEAN
    else:
        return str(value), xlrd.XL_CELL_TEXT
    return value, dt

class Base(object):

    def __init__(self, sheet_writer, cell_node, value, data_type):
        self.sheet_writer = sheet_writer
        self.cell_node = cell_node
        self.value = value
        self.data_type = data_type

    @property
    def rdsheet(self):
        return self.sheet_writer.rdsheet

    @property
    def wtsheet(self):
        return self.sheet_writer.wtsheet

    @property
    def source_cell(self):
        return self.cell_node.sheet_cell

    @property
    def target_cell(self):
        return None

    @property
    def rdcolx(self):
        return self.cell_node.colx

    @property
    def rdrowx(self):
        return self.cell_node.rowx

    @property
    def wtcolx(self):
        return self.sheet_writer.box.right

    @property
    def wtrowx(self):
        return self.sheet_writer.box.bottom

    def apply_filters(self):
        if hasattr(self.cell_node, 'filters') and self.cell_node.filters:
            for (filter, args) in self.cell_node.filters:
                #print('args', args)
                filter(self, *args)
            self.cell_node.filters.clear()


class CellContextX(Base):

    def __init__(self, sheet_writer, cell_node, value, data_type):
        super().__init__(sheet_writer, cell_node, value, data_type)
        self._target_cell = None

    @property
    def target_cell(self):
        if self._target_cell:
            return self._target_cell
        source = self.source_cell
        wtcolx = self.wtcolx
        wtrowx = self.wtrowx
        value = self.value
        data_type = self.data_type
        target = self.wtsheet.cell(column=wtcolx, row=wtrowx)
        if value is None:
            target._value = source._value
            target.data_type = source.data_type
        elif isinstance(value, STRING_TYPES) and value.startswith('='):
            target.value = value
        elif data_type:
            target._value = value
            target.data_type = data_type
        else:
            target.value = value
        if source.has_style:
            target._style = copy.copy(source._style)
        if source.hyperlink:
            target.hyperlink = copy.copy(source.hyperlink)
        self._target_cell = target
        return self._target_cell

    def get_style(self):
        return self.target_cell.style

    def finish(self):
        self.apply_filters()



class CellContext(Base):

    def __init__(self, sheet_writer, cell_node, value, data_type):
        super().__init__(sheet_writer, cell_node, value, data_type)
        self._style = None

    def get_style(self):
        if self._style:
            return self._style
        source_cell = self.source_cell
        if source_cell.xf_index is not None:
            style = self.sheet_writer.style_list[source_cell.xf_index]
        else:
            style = self.style_list[0]
        self._style = copy.copy(style)
        return self._style

    def set_cell(self):
        cty = self.data_type
        if cty == xlrd.XL_CELL_EMPTY:
            return
        source_cell = self.source_cell
        rdrowx = self.rdrowx
        rdcolx = self.rdcolx
        wtrowx = self.wtrowx
        wtcolx = self.wtcolx
        value = self.value
        style = self.get_style()

        if value is None:
            value = source_cell.value
            cty = source_cell.ctype
        if cty is None:
            value, cty = get_type(value)

        wtrow = self.wtsheet.row(wtrowx)
        if cty == xlrd.XL_CELL_TEXT:
            if isinstance(value, (list, tuple)):
                wtrow.set_cell_rich_text(wtcolx, value, style)
            elif value.startswith('='):
                try:
                    formula = xlwt.Formula(value[1:])
                    wtrow.set_cell_formula(wtcolx, formula, style)
                except BaseException as e:
                    wtrow.set_cell_text(wtcolx, value, style)
            else:
                wtrow.set_cell_text(wtcolx, value, style)
        elif cty == xlrd.XL_CELL_NUMBER or cty == xlrd.XL_CELL_DATE:
            wtrow.set_cell_number(wtcolx, value, style)
        elif cty == xlrd.XL_CELL_BLANK:
            wtrow.set_cell_blank(wtcolx, style)
        elif cty == xlrd.XL_CELL_BOOLEAN:
            wtrow.set_cell_boolean(wtcolx, value, style)
        elif cty == xlrd.XL_CELL_ERROR:
            wtrow.set_cell_error(wtcolx, value, style)
        else:
            raise Exception(
                "Unknown xlrd cell type %r with value %r at (sheet=%r,rowx=%r,colx=%r)" \
                % (cty, value, self.rdsheet.name, rdrowx, rdcolx)
            )
    def finish(self):
        self.apply_filters()
        self.set_cell()
